"""Utility code to use in all other modules."""
from .accept import TVisitor, accept, on_unknown_acceptor
from .io import MCNP_ENCODING, check_if_all_paths_exist, check_if_path_exists, make_dir, make_dirs
from .misc import (
    MAX_DIGITS,
    are_equal,
    compute_hash,
    deepcopy,
    filter_dict,
    get_decades,
    is_in,
    is_sorted,
    make_hashable,
    mids,
    prettify_float,
    round_array,
    round_scalar,
    significant_array,
    significant_digits,
)
from .resource import filename_resolver, path_resolver
from .tolerance import FLOAT_TOLERANCE
